function mex(cudafile,varargin)
% mex(cudafile) Builds MEX function from CUDA code using (or not) Thrust API.
% mex(___,Name,Value)  specifies options using one or more option name, option 
%   value pairs. Name is the option name and must appear inside single quotes 
%   (''). Value is the corresponding value. 
% Here are the different options :
% mex(___,'outpath',Value), specifies where the MEX file will be written.
%   Value is a string (current directory as default)
% mex(___,'force',Value), specifies if compilation is forced or not.
%   Value is a logical (false as default)
% mex(___,'samepath',Value), specifies if the MEX file will be written in the 
%   same path as input file. Value is a logical (false as default)
% mex(___,'autopath',Value), specifies if the MEX file will be written in a 
%   directory automaticaly selected. Value is a logical (false as default)
% mex(___,'dep',Value), specifies dependencies for Makefile
%   Value is a string (empty as default)
  p = inputParser; 
  p.addParamValue('verbose', 1, @(x) ismember(x,[0:2]) );
  p.addParamValue('outpath', '.', @ischar );
  p.addParamValue('force', false, @islogical ); % force compilation if true
  p.addParamValue('samepath', false, @islogical );
  p.addParamValue('autopath', false, @islogical );
  p.addParamValue('debug', false, @islogical );
  p.addParamValue('dep', '', @ischar ); % dependance for the Makefile
  p.parse(varargin{:});
  R=p.Results;
  verbose=p.Results.verbose;
  if ~strcmp('.cu',cudafile(end-2:end))
    error('Input file must be a .cu file!')
  end
  toolboxdir=Thrust.tools.getToolboxpath();
  if R.autopath
    if R.samepath
      R.samepath=false;
      fprintf('-> force ''samepath'' option to false\n');
    end
    P=Thrust.tools.getConfigPaths();
    if R.debug
      R.outpath=P.MEXDEBUG_DIR;
    else
      R.outpath=P.MEX_DIR;
    end
    if exist(R.outpath)~=7,mkdir(R.outpath);end
    if verbose>0,
      fprintf('[Matlab] set ''outpath'' to:\n  %s\n',R.outpath);
    end
  end
  if R.samepath
    I=strfind(cudafile,filesep);
    if ~isempty(I)
      R.outpath=cudafile(1:I(end)-1);% suppress name and last filesep
      if verbose>0,
        fprintf('[Matlab] set ''outpath'' to:\n  %s\n',R.outpath);
      end
    end
  end
  
  mexfile=BuildMakefile(toolboxdir,cudafile,R);
  if verbose>0
    fprintf('[Matlab] Build file Makefile.mThrust\n');
    fprintf('[Matlab] Run system command "make -f Makefile.mThrust"\n');
    fprintf('    That can take a few minutes ... be patient\n');
  end
  if (R.force) & (exist(mexfile)==2)
    delete(mexfile)
  end
  command='make --no-print-directory -f Makefile.mThrust';
  [status,result]=system(command);
  if status~=0
    fprintf('Compilation failed:\n')
    fprintf('  -> %s\n',command);
    fprintf('See Makefile.log\n');
    fid=fopen('Makefile.log','w');
    fprintf(fid,result)
    fclose(fid);
    if verbose>0
      %Thrust.extern.cprintf.cprintf('Red','%s',result);
      fprintf('%s\n',result);
    end
    error('Compilation failed!');
  elseif verbose>0
    %Thrust.extern.cprintf.cprintf('Blue','%s',result)
    fprintf('%s\n',result);
  end
end

function mexfile=BuildMakefileold(toolboxdir,cudafile,R)
  fid = fopen('Makefile.mThrust','w');
  fprintf(fid,'# automaticaly generated by mThrust_mex Matlab function \n');
  fprintf(fid,'MTHRUST_DIR = %s\n',toolboxdir);
  fprintf(fid,'MAIN_DIR = $(MTHRUST_DIR)\n');
  fprintf(fid,'include $(MTHRUST_DIR)/Makefile.inc\n');
  fprintf(fid,'include $(BUILD_DIR)/config\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.matlab\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.cuda-$(CUDA_VERSION)\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.mex\n\n');
  fprintf(fid,'NVCC_FLAGS += -Xcompiler ''-fPIC  -O3 -funroll-loops -msse3'' -D FAST_MATH=1 -O3 -m 64 --use_fast_math $(GENCODE_FLAGS)\n\n');
  fprintf(fid,'INCLUDELIB = -L$(CUDA_PATH)/lib64  -lcudart  $(CUDASDK_LIB)\n\n');
  
  fprintf(fid,'.SUFFIXES: .$(MEXEXT) .cu\n\n');
  I=strfind(cudafile,'.');
  J=strfind(cudafile,filesep);
  if isempty(J)
    cudaname=cudafile(1:I(end)-1); 
  else
    cudaname=cudafile(J(end)+1:I(end)-1); % without path and extension
  end
  if R.verbose>1, echo='';else echo='@';end
  %fprintf(fid,'%s.$(MEXEXT) : %s %s\n',[R.outpath,filesep,cudaname],strrep(cudafile,'.cu','.o'),strrep(R.dep,'.cu','.o')); % to improve
  fprintf(fid,'%s.$(MEXEXT) : %s\n',[R.outpath,filesep,cudaname],strrep(cudafile,'.cu','.o')); % to improve
  fprintf(fid,'\t@echo "********** MEX  compiler *************"\n');
  fprintf(fid,'\t@echo "MEX compiler used to build $@ ..."\n');
  fprintf(fid,'\t%s$(MEX) $(MEXFLAGS)  -cxx $(INCLUDELIB) $(CUDASDK_LIB) $^ -L$(MEX_DIR) -lmThrust -lc -output $@\n',echo);
  %fprintf(fid,'\t@rm $*.o\n\n');
  fprintf(fid,'\n%s : %s %s\n\n',strrep(cudafile,'.cu','.o'),cudafile,R.dep);
  fprintf(fid,'\n%%.o : %%.cu\n');
  fprintf(fid,'\t@echo "********** NVCC compiler *************"\n');
  fprintf(fid,'\t@echo "NVCC compiler used to build $*.o ..."\n');
  fprintf(fid,'\t%s$(NVCC) $(NVCC_FLAGS) -I$(INC_DIR) -I$(MATLAB_INC) -I$(SRC_AUTO_DIR) -c $< -o $*.o\n\n',echo);
  fclose(fid);
  mexfile=[R.outpath,filesep,cudaname,'.',mexext];
end

function mexfile=BuildMakefile(toolboxdir,cudafile,R)
  fid = fopen('Makefile.mThrust','w');
  fprintf(fid,'# automaticaly generated by Thrust.compile.mex Matlab function \n');
  fprintf(fid,'MTHRUST_DIR = %s\n',toolboxdir);
  fprintf(fid,'MAIN_DIR = $(MTHRUST_DIR)\n');
  fprintf(fid,'include $(MTHRUST_DIR)/Makefile.inc\n');
  fprintf(fid,'include $(BUILD_DIR)/config\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.matlab\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.cuda-$(CUDA_VERSION)\n');
  fprintf(fid,'include $(BUILD_DIR)/Makefile.mex\n\n');
  fprintf(fid,'NVCC_FLAGS += -Xcompiler ''-fPIC  -O3 -funroll-loops -msse3'' -D FAST_MATH=1  -D_FORCE_INLINES -O3 -m 64 --use_fast_math $(GENCODE_FLAGS)\n\n');
  fprintf(fid,'INCLUDELIB = -L$(CUDA_PATH)/lib64  -lcudart  $(CUDASDK_LIB)\n\n');
  fprintf(fid,'DEBUGFLAGS =\n');
  if R.debug
    fprintf(fid,'DEBUGFLAGS += -D DEBUG=1\n');
    fprintf(fid,'DEBUGFLAGS += -D VERBOSE=1\n');
    fprintf(fid,'DEBUGFLAGS += -D DEBUG_GPUMEMORY=1\n');
  end
  
  fprintf(fid,'.SUFFIXES: .$(MEXEXT) .cu\n\n');
  I=strfind(cudafile,'.');
  J=strfind(cudafile,filesep);
  if isempty(J)
    cudaname=cudafile(1:I(end)-1); 
  else
    cudaname=cudafile(J(end)+1:I(end)-1); % without path and extension
  end
  if R.verbose>1, echo='';else echo='@';end
  %fprintf(fid,'%s.$(MEXEXT) : %s %s\n',[R.outpath,filesep,cudaname],strrep(cudafile,'.cu','.o'),strrep(R.dep,'.cu','.o')); % to improve
  fprintf(fid,'%s.$(MEXEXT) : %s\n',[R.outpath,filesep,cudaname],cudafile); % to improve
  fprintf(fid,'\t@echo "********** NVCC compiler *************"\n');
  fprintf(fid,'\t@echo "NVCC compiler used to build $*.o ..."\n');
  fprintf(fid,'\t%s$(NVCC) $(DEBUGFLAGS) $(NVCC_FLAGS) -I$(INC_DIR) -I$(MATLAB_INC) -I$(SRC_AUTO_DIR) -c $< -o $*.o\n',echo);
  fprintf(fid,'\t@echo "********** MEX  compiler *************"\n');
  fprintf(fid,'\t@echo "MEX compiler used to build $@ ..."\n');
  fprintf(fid,'\t%s$(MEX) $(MEXFLAGS)  -cxx $(INCLUDELIB) $(CUDASDK_LIB) $*.o -L$(MEX_DIR) -lmThrust -lc -output $@\n',echo);
  fprintf(fid,'\t@rm $*.o\n');
  fclose(fid);
  mexfile=[R.outpath,filesep,cudaname,'.',mexext];
end