function varargout=GeneralizedNewtonFractalGPU(sf,sfp,z0,varargin);
% GeneralizedNewtonFractalGPU solves a nonlinear system of equations
%   Z=GeneralizedNewtonFractalVec(sf,sfp,z0) 
%   solves the nonlinear system of equations f(z)=0 using 
%   Newtons methods with z real or complex, starting with 
%   the initial guess z0 up to a default tolerance 1e-8, 
%   doing at most 200 iterations. 
%   The analytical derivative of f is given by the 
%   parameter sfp. The initial guess z0 could be an array. 
%   (vectorized Matlab version).
  import Thrust.samples.NewtonFractal.*;
  p = inputParser;
  p.addParamValue('tol', 1e-8, @isscalar );
  p.addParamValue('itermax', 200, @isscalar );
  p.addParamValue('verbose', 0, @(x) ismember(x,0:1) );
  p.addParamValue('a', 1, @isscalar );
  p.addParamValue('compile', true, @islogical );
  p.parse(varargin{:});
  R=p.Results;
  a=p.Results.a;
  verbose=p.Results.verbose;
  tol=p.Results.tol;
  itermax=p.Results.itermax;
  
  if R.compile
    if R.verbose,fprintf('0) CUDA/Mex compilation : ');end
    P=Thrust.tools.getCurrentFunctionPath();
    tstart=tic();
    file=create_funcs(sf,sfp,'file','funcs.cuh');
    cudamexfile=[P,'mex_GeneralizedNewtonFractal_full.cu'];
    dep=[P,'funcs.cuh'];
    Thrust.compile.mex(cudamexfile,'samepath',true,'verbose',0,'dep',dep)
    T(1)=toc(tstart);
    if R.verbose,fprintf('%.3f(s)\n',T(1));end
  end
  if R.verbose,fprintf('1) Transfert/init datas on GPU\n');end
  tstart=tic();
  gz0=mThrust(z0(:),'device'); % Only vectors are supported on mThrust
  gz=mThrust(numel(z0),'device','double',true);
  gcount=mThrust(numel(z0),'device','unsigned',false);
  cuda.DeviceSynchronize();
  T(2)=toc(tstart);
  if R.verbose,fprintf('2) Computation on GPU\n');end
  tstart=tic();
  mex_GeneralizedNewtonFractal_full(complex(a),gz0,uint32(itermax),tol,gz,gcount);
  cuda.DeviceSynchronize();
  T(3)=toc(tstart);
  if R.verbose,fprintf('3) Transfert datas on Matlab\n');end
  tstart = tic();
  z=reshape(gather(gz),size(z0));
  count=reshape(gather(gcount),size(z0));
  cuda.DeviceSynchronize();
  T(4) = toc( tstart ); % transfert time to CPU
  if R.verbose
  fprintf('   to GPU : %.3f(s), computation: %.3f(s), to CPU: %.3f(s)\n',T(2:4))
  fprintf('  -> total: %.3f(s)\n',sum(T(2:4)));
  end
  if nargout>=1,varargout{1}=z;end
  if nargout>=2,varargout{2}=count;end
  if nargout==3,varargout{3}=T;end
end