function varargout=GeneralizedNewtonFractalVec(f,fp,z0,varargin);
% GeneralizedNewtonFractalVec solves a nonlinear system of equations
%   Z=GeneralizedNewtonFractalVec(f,fp,z0) 
%   solves the nonlinear system of equations f(z)=0 using 
%   Newtons methods with z real or complex, starting with 
%   the initial guess z0 up to a default tolerance 1e-8, 
%   doing at most 200 iterations. 
%   The analytical derivative of f is given by the 
%   parameter fp. The initial guess z0 could be an array. 
%   (vectorized Matlab version).
  p = inputParser;
  p.addParamValue('tol', 1e-8, @isscalar );
  p.addParamValue('itermax', 200, @isscalar );
  p.addParamValue('verbose', 0, @(x) ismember(x,0:1) );
  p.addParamValue('a', 1, @isscalar );
  p.parse(varargin{:});
  a=p.Results.a;
  verbose=p.Results.verbose;
  tol=p.Results.tol;
  itermax=p.Results.itermax;
  tstart=tic();
  k=1;err=tol+1;alpha=[];
  gz0=gpuArray(z0(:)); 
  gz=gz0;
  gcount=gpuArray.zeros(numel(z0),1,'uint32');
  T(1)=toc(tstart);
  wait(gpuDevice)
  tstart=tic();
  bconv=false; % logical true if convergence
  while ( (~bconv) & (k<=itermax) )
    dz=a*f(gz)./fp(gz);
    gz=gz-dz; 
    notconv=uint32(abs(dz)>tol);
    gcount = gcount + notconv;
    bconv = (sum(notconv)==0);
    k=k+1;
    if verbose>0,fprintf('k=%d\n',k);end
  end
  wait(gpuDevice)
  T(2)=toc(tstart);
  z=reshape(gather(gz),size(z0));
  count=reshape(gather(gcount),size(z0));
  wait(gpuDevice)
  T(3)=toc(tstart);
  if nargout>=1,varargout{1}=z;end
  if nargout>=2,varargout{2}=count;end
  if nargout==3,varargout{3}=T;end
end