#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <thrust/complex.h>

__device__ thrust::complex<double> f(thrust::complex<double> z){
  return  z*z*z-1.0;
}
__device__ thrust::complex<double> df(thrust::complex<double> z){
  return  3.0*z*z;
}
//
// A generalization of Newton's iteration is :
// z_{n+1} = z_{n} - a f(z_{n})/f'(z_{n})
//
__device__ void GeneralizedNewtonFractal( 
                          const thrust::complex<double> &a,
                          const thrust::complex<double> &z0,
                          const unsigned int maxIters, 
                          const double tol,
                          thrust::complex<double> &z,
                          unsigned int &count
                              ) {
  bool bconv=false;
  thrust::complex<double> dz;
  count=0;
  z=z0;
  double tol2=tol*tol;
  while ( ( count <= maxIters ) && (!bconv) ){
      count++;
      dz=a*f(z)/df(z);
      z=z-dz;
      bconv= dz.real()*dz.real()+dz.imag()*dz.imag() <= tol2;
      //more efficient than bconv= (abs(dz) <= tol);
  }
}

__global__ void GeneralizedNewtonFractal_kernel(
                      const thrust::complex<double> a,
                      const thrust::complex<double> *z0, 
                      const unsigned int numel,
                      const unsigned int maxIters, 
                      const double tol,
                      thrust::complex<double> *z,
                      unsigned int *count)
{
  const unsigned int      tid = blockDim.x * blockIdx.x + threadIdx.x;
  const unsigned int THREAD_N = blockDim.x * gridDim.x;
  unsigned int globalThreadIdx=tid;
  while (globalThreadIdx < numel){
    GeneralizedNewtonFractal(a, z0[globalThreadIdx], maxIters,tol,z[globalThreadIdx],count[globalThreadIdx]);
    globalThreadIdx+=THREAD_N;
  }
}

// See https://devblogs.nvidia.com/parallelforall/cuda-pro-tip-occupancy-api-simplifies-launch-configuration/
// Needs at least CUDA 6.5
void getKernelSizes(int &gridSize, int &blockSize, int arraysize){
  int minGridSize;
  cudaOccupancyMaxPotentialBlockSize(&minGridSize, &blockSize, GeneralizedNewtonFractal_kernel, 0, arraysize); 
  gridSize = (arraysize + blockSize - 1) / blockSize;
}

// prhs[0] : a Matlab complex double
// prhs[1] : z0 mThrust complex double
// prhs[2] : maxIter unsigned
// prhs[3] : tol double
// prhs[4] : z mThrust complex double (output)
// prhs[5] : count mThrust unsigned (output)
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 6) 
    mexErrMsgTxt("Bad input.");
  typedef thrust::complex<double>              Cplx;
  typedef thrust::device_vector<Cplx>          VecCplx;
  typedef thrust::device_vector<unsigned int>  VecInt;
  
  double *pa_r = (double *) mxGetPr(prhs[0]);
  double *pa_i = (double *) mxGetPi(prhs[0]);
  Cplx a=Cplx(*pa_r,*pa_i);
  VecCplx &z0   = get_object< VecCplx >(mxGetProperty(prhs[1],0,"handle"));
  unsigned int *pmaxIter = (unsigned int *) mxGetPr(prhs[2]);
  double *ptol = (double *) mxGetPr(prhs[3]);
  VecCplx &z    = get_object< VecCplx >(mxGetProperty(prhs[4],0,"handle"));
  VecInt &count = get_object< VecInt >(mxGetProperty(prhs[5],0,"handle"));
  
  Cplx *pz0=(Cplx *)thrust::raw_pointer_cast(&z0[0]);
  Cplx *pz =(Cplx *)thrust::raw_pointer_cast(&z[0]);
  unsigned int *pcount=(unsigned int *)thrust::raw_pointer_cast(&count[0]);
  
  int blockSize; // The launch configurator returned block size 
  int gridSize;  // The actual grid size needed, based on input size 
  getKernelSizes(gridSize,blockSize, z0.size());
  
  GeneralizedNewtonFractal_kernel<<<gridSize,blockSize>>>(a, pz0, z0.size() ,*pmaxIter, *ptol, pz, pcount);
  MEXPRINTF("ending");
}