/*! \file mex_mThrust_create.cu
 *  \brief MEX-file : create a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayCreate create a C++ \p gTArray object and 
 *  return handle to this object so that it's possible to use it in another MEX-file.
 * 
 *  &copy;Matlab syntax <tt><B>h=mex_gTArrayCreate(N,type)</B></tt> where
 *    - \p N  size of \p gTArray (must be \p int32)
 *    - \p type must be 'int', 'unsigned', 'float' or 'double'
 *    - \p h is a handle to created C++ \p gTArray object.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayCreate 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  \endcode
 */

#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"

#define MESS "mThrustToolbox:create"

template  <typename T>
mxArray *mx_mThrust_create_type(char *container,char *type,int N,logical iscomplex){
  if (strcmp(container, "host")==0){
    if (!iscomplex){
      thrust::host_vector<T> *pV = new thrust::host_vector<T>(N);
      PRINT_DEBUGSTR("create",container,type,pV->size(),pV);
      ObjectHandle<thrust::host_vector<T> > *handle = new ObjectHandle<thrust::host_vector<T> >(*pV);
      return handle->to_mex_handle();
    }else{
      thrust::host_vector<thrust::complex<T> > *pV = new thrust::host_vector<thrust::complex<T> >(N);
      PRINT_DEBUGSTRC("create",container,type,pV->size(),pV);
      ObjectHandle<thrust::host_vector<thrust::complex<T> > > *handle = new ObjectHandle<thrust::host_vector<thrust::complex<T> > >(*pV);
      return handle->to_mex_handle();
    }
  }else if (strcmp(container, "device")==0){
    if (!iscomplex){
      thrust::device_vector<T> *pV = new thrust::device_vector<T>(N);
      PRINT_DEBUGSTR("create",container,type,pV->size(),pV);
      ObjectHandle<thrust::device_vector<T> > *handle = new ObjectHandle<thrust::device_vector<T> >(*pV);
      return handle->to_mex_handle();
    }else{
      thrust::device_vector<thrust::complex<T> > *pV = new thrust::device_vector<thrust::complex<T> >(N);
      PRINT_DEBUGSTRC("create",container,type,pV->size(),pV);
      ObjectHandle<thrust::device_vector<thrust::complex<T> > > *handle = new ObjectHandle<thrust::device_vector<thrust::complex<T> > >(*pV);
      return handle->to_mex_handle();
    }
  }else
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
  return 0;
}

/*! The gateway Cuda/C++ function \p mex_gTArrayCreate create a C++ \p gTArray object and 
 *  return handle to this object so that it's possible to use it in another MEX-file.
 * 
 *  With \p nlhs==1 and  \p nrhs==2, the Matlab arguments are
 *    - \p prhs[0]  size of \p gTArray (must be \p int32)
 *    - \p prhs[1] must be 'int', 'unsigned', 'float' or 'double'
 *    - \p plhs[0] is a handle to created C++ \p gTArray object.
 *  
 *  \param[out] nlhs Number of expected output mxArray(s).
 *  \param[out] plhs Array of pointers to the expected output mxArray (s).
 *  \param[in]  nrhs Number of input mxArray (s).
 *  \param[in]  prhs Array of pointers to the input mxArray (s).
 * 
 *  \see http://www.mathworks.fr/fr/help/matlab/matlab_external/building-mex-files.html
 *  \see http://www.mathworks.fr/fr/help/matlab/apiref/mexfunction.html
 */
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  bool Lock=mexIsLocked();
  if (nlhs == 0 || nrhs == 0){
    if (Lock){
      mexUnlock();
      MEXVERBOSE("Unlock mex file");
//       #ifdef VERBOSE
//       mexPrintf("%sUnlock mex file : %s...%s\n",VERBOSECOLOR,__FILE__,TEXTCOLOR_END);
//       #endif
    }
    return ;
  }else if (nlhs !=1 || nrhs != 4) 
    mexErrMsgTxt("Bad input... To do");
  
  if (!mxIsLogicalScalar(prhs[3]))
    mexErrMsgIdAndTxt(MESS,"4th onput value must be a logical.");
  
  if( !mxIsInt32(prhs[0]) || mxGetNumberOfElements(prhs[0])!=1 ) {
    mexErrMsgIdAndTxt(MESS,"Input value must be an int32.");
  }
  
  if (!Lock){
    mexLock();
    MEXVERBOSE("Lock mex file");
//     #ifdef VERBOSE      
//     mexPrintf("%sLock mex file : %s...%s\n",VERBOSECOLOR,__FILE__,TEXTCOLOR_END);
//     #endif
  }  
  
  logical *pisComplex=(logical *) mxGetPr(prhs[3]);
  int * pn = (int *) mxGetPr(prhs[0]);
  char *type =mxArrayToString(prhs[2]);
  char *container =mxArrayToString(prhs[1]);
  
  if (strcmp(type, "int")==0)
    plhs[0]=mx_mThrust_create_type<int>(container,type,*pn,*pisComplex);
  else if (strcmp(type, "double")==0)
    plhs[0]=mx_mThrust_create_type<double>(container,type,*pn,*pisComplex);
  else if (strcmp(type, "float")==0)
    plhs[0]=mx_mThrust_create_type<float>(container,type,*pn,*pisComplex);
  else if (strcmp(type, "unsigned")==0)
    plhs[0]=mx_mThrust_create_type<unsigned>(container,type,*pn,*pisComplex);
  else if (strcmp(type, "logical")==0)
    plhs[0]=mx_mThrust_create_type<logical>(container,type,*pn,*pisComplex);
  else{
    mxFree(type);mxFree(container);
    mexErrMsgIdAndTxt(MESS,"Type not implemented.");
  }
  mxFree(type);mxFree(container);
  MEXPRINTF("ending");
}