/*! \file mex_gTArrayDelete.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <thrust/device_ptr.h>

#define MESS "mThrustToolbox:delete"

template <typename T,typename Vector> 
void mx_mThrust_op(const mxArray *prhs[],char *container,char *type,logical iscomplex){
  Vector &pV = get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  if (iscomplex){
    PRINT_DEBUGSTRC("delete",container,type,pV.size(),&pV);
//     thrust::complex<T> *pz0=(thrust::complex<T> *)thrust::raw_pointer_cast(&pV[0]);
//     thrust::device_ptr<thrust::complex<T> > toto=thrust::device_pointer_cast(pz0);
//     thrust::device_free(toto);
    //cudaFree((void *)pz0);
  }else{
    PRINT_DEBUGSTR("delete",container,type,pV.size(),&pV);
//     T *pz0=(T *)thrust::raw_pointer_cast(&pV[0]);
//     thrust::device_ptr<T> toto=thrust::device_pointer_cast(pz0);
//     thrust::device_free(toto);
    //cudaFree((void *)pz0);
  }
  //thrust::complex<double> *pz0=(thrust::complex<double> *)thrust::raw_pointer_cast(&z0[0]);
  delete &pV;
  //pV.clear();// empty the vector
  // deallocate any capacity which may currently be associated with pV
  //pV.shrink_to_fit();
  //Vector().swap(pV);
}

template  <typename T>
void mx_mThrust_delete_type(const mxArray *prhs[],char *container,char *type,logical iscomplex){
  if (strcmp(container, "host")==0){
    if (!iscomplex){
      mx_mThrust_op<T,thrust::host_vector<T> >(prhs,container,type,iscomplex);
    }else{
      mx_mThrust_op<T,thrust::host_vector<thrust::complex<T> > >(prhs,container,type,iscomplex);
    }
  }else if (strcmp(container, "device")==0){
    if (!iscomplex){
      mx_mThrust_op<T,thrust::device_vector<T> >(prhs,container,type,iscomplex);
    }else{
      mx_mThrust_op<T,thrust::device_vector<thrust::complex<T> > >(prhs,container,type,iscomplex);
    }
  }else
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
}

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting"); 
//   bool Lock=mexIsLocked();
//   if (!Lock){
//     mexLock();
//     #ifdef VERBOSE      
//     mexPrintf("%sLock mex file : %s...%s\n",VERBOSECOLOR,__FILE__,TEXTCOLOR_END);
//     #endif
//   }  
  if (nlhs !=0 || nrhs != 1) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust"));
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  char *container =mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  logical *pisComplex=(logical *) mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  
  if (strcmp(type, "double")==0)
    mx_mThrust_delete_type<double>(prhs,container,type,*pisComplex);
  else if (strcmp(type, "int")==0)
    mx_mThrust_delete_type<int>(prhs,container,type,*pisComplex);
  else if (strcmp(type, "float")==0)
    mx_mThrust_delete_type<float>(prhs,container,type,*pisComplex);
  else if (strcmp(type, "unsigned")==0)
    mx_mThrust_delete_type<unsigned>(prhs,container,type,*pisComplex);
  else if (strcmp(type, "logical")==0)
    mx_mThrust_delete_type<logical>(prhs,container,type,*pisComplex);
  else{
    mxFree(type);mxFree(container);
    mexErrMsgIdAndTxt(MESS,"Type not implemented.");
  }
  mxFree(type);mxFree(container);
  MEXPRINTF("ending");
}