/*! \file mex_mThrust_disp.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <iostream>
#include <iomanip>
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <thrust/complex.h>
 
#define MESS "mThrustToolbox:disp"

// template <typename T,typename Vector> void  mThrust_disp_opCplx(const mxArray *prhs[],
//                                                             int N,char *container,char *type,int precision){
//   Vector &Vr = get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
//   if (Vr.size() == 0)
//     std::cout << "Empty Thrust vector\n";
//   else{
//     std::cout << " thrust::"<<container<<"_vector<";
//     std::cout <<"complex<"<<type<<">>[" << Vr.size() << "]\n ";
//     std::cout << std::setprecision(precision);
//     if (2*N>=Vr.size()){
//       int n=(N<Vr.size()?N:Vr.size());
//       for (int i=0;i<n;i++)
//         std::cout <<  Vr[i] << "  ";
//       if (n<Vr.size()){
//         std::cout << std::endl;
//         for (int i=n;i<Vr.size();i++)
//           std::cout <<  Vr[i] << "  ";
//       }
//       std::cout << std::endl;
//     }else{
//       for (int i=0;i<N;i++)
//         std::cout <<  Vr[i] << "  ";
//       std::cout << "\n . . .\n ";
//       for (int i=Vr.size()-N;i<Vr.size();i++)
//         std::cout <<  Vr[i] << "  ";
//       std::cout << "\n";
//     }
//   }
// }


#if defined THRUST_VERSION <= 100801
#define PRINTCPLX(z) std::cout << "(" << z.real() << "," << z.imag() << ")"
#else
#define PRINTCPLX(z) std::cout << z
#endif

                                                            
template <typename T,typename Vector> void  mThrust_disp_op_complex(const mxArray *prhs[],
                                                            int N,char *container,char *type,int precision){
  Vector &Vr = get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  if (Vr.size() == 0)
    std::cout << "Empty Thrust vector\n";
  else{
    std::cout << " thrust::"<<container<<"_vector<";
    std::cout <<"complex<"<<type<<">>[" << Vr.size() << "]\n ";
    std::cout << std::setprecision(precision);
    if (2*N>=Vr.size()){
      int n=(N<Vr.size()?N:Vr.size());
      for (int i=0;i<n;i++){
	thrust::complex<T> z=Vr[i];
	std::cout << "(" << z.real() << "," << z.imag() << ")";
        std::cout << "  ";
      }
      if (n<Vr.size()){
        std::cout << std::endl;
        for (int i=n;i<Vr.size();i++){
	  thrust::complex<T> z=Vr[i];
	  std::cout << "(" << z.real() << "," << z.imag() << ")";
	  std::cout << "  ";
        }  
      }
      std::cout << std::endl;
    }else{
      for (int i=0;i<N;i++){
	thrust::complex<T> z=Vr[i];
	std::cout << "(" << z.real() << "," << z.imag() << ")";
        std::cout << "  ";
      }
      std::cout << "\n . . .\n ";
      for (int i=Vr.size()-N;i<Vr.size();i++){
	thrust::complex<T> z=Vr[i];
	std::cout << "(" << z.real() << "," << z.imag() << ")";
	std::cout << "  ";
      }
      std::cout << "\n";
    }
  }
}

template <typename T,typename Vector> void  mThrust_disp_op_real(const mxArray *prhs[],
                                                            int N,char *container,char *type,int precision){
  Vector &Vr = get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  if (Vr.size() == 0)
    std::cout << "Empty Thrust vector\n";
  else{
    std::cout << " thrust::"<<container<<"_vector<";
    std::cout <<type<<">[" << Vr.size() << "]\n ";
    std::cout << std::setprecision(precision);
    if (2*N>=Vr.size()){
      int n=(N<Vr.size()?N:Vr.size());
      for (int i=0;i<n;i++)
	std::cout << Vr[i] << "  ";
      if (n<Vr.size())
        std::cout << std::endl;
        for (int i=n;i<Vr.size();i++)
          std::cout << Vr[i] << "  ";
      std::cout << std::endl;
    }else{
      for (int i=0;i<N;i++) std::cout << Vr[i] << "  ";
      std::cout << "\n . . .\n ";
      for (int i=Vr.size()-N;i<Vr.size();i++) std::cout << Vr[i] << "  ";
	
      std::cout << "\n";
    }
  }
}


template  <typename T>
void mx_mThrust_disp_type(const mxArray *prhs[],char *container,char *type,logical iscomplex,int N,int precision){
  if (strcmp(container, "host")==0){
    if (iscomplex){
      mThrust_disp_op_complex<T,thrust::host_vector<thrust::complex<T> > >(prhs,N,container,type,precision);
    }else{
      mThrust_disp_op_real<T,thrust::host_vector<T> >(prhs,N,container,type,precision);
    }
  }else if (strcmp(container, "device")==0){
    if (iscomplex){
      mThrust_disp_op_complex<T,thrust::device_vector<thrust::complex<T> > >(prhs,N,container,type,precision);
    }else{
      mThrust_disp_op_real<T,thrust::device_vector<T> >(prhs,N,container,type,precision);
    }
  }else
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
}
// prhs[0] -> mThrust object
// prhs[1] -> number to print (int32)
// prhs[2] -> digit precision (int32)
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 3) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust"));
  logical *pisComplex=(logical *) mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  if( !mxIsInt32(prhs[1]) || mxGetNumberOfElements(prhs[1])!=1 )
    mexErrMsgIdAndTxt("mThrust:disp","2nd input value must be an int32.");
  if( !mxIsInt32(prhs[2]) || mxGetNumberOfElements(prhs[2])!=1 ){
    printf("Number of Elem : %d\n",mxGetNumberOfElements(prhs[2]));
    mexErrMsgIdAndTxt("mThrust:disp","3rd input value must be an int32.");
  }
  int *pn = (int *) mxGetPr(prhs[1]);
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  char *container =mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  int *pprecision = (int *) mxGetPr(prhs[2]);
  
  if (strcmp(type, "int")==0)
    mx_mThrust_disp_type<int>(prhs,container,type,*pisComplex,*pn,*pprecision);
  else if (strcmp(type, "double")==0)
    mx_mThrust_disp_type<double>(prhs,container,type,*pisComplex,*pn,*pprecision);
  else if (strcmp(type, "float")==0)
    mx_mThrust_disp_type<float>(prhs,container,type,*pisComplex,*pn,*pprecision);
  else if (strcmp(type, "unsigned")==0)
    mx_mThrust_disp_type<unsigned>(prhs,container,type,*pisComplex,*pn,*pprecision);
  else if (strcmp(type, "logical")==0)
    mx_mThrust_disp_type<logical>(prhs,container,type,*pisComplex,*pn,*pprecision);
  else{ 
    mxFree(container);mxFree(type);
    mexErrMsgIdAndTxt(MESS,"Type not implemented.");
  }
  mxFree(container);mxFree(type);
  MEXPRINTF("ending");
}