/*! \file mex_mThrust_disp.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <iostream>
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include "complex_functors.h"

template <typename T, typename Vector> 
void complexCopyDevice(mxArray *plhs[],const mxArray *prhs[], mxClassID classid){
  Vector &V=get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  mwSize dims0[1];
  dims0[0]=V.size();
  plhs[0] = mxCreateNumericArray(1,dims0,classid,mxCOMPLEX);
  T *pr= (T *)mxGetPr(plhs[0]);
  T *pi= (T *)mxGetPi(plhs[0]);
  
  thrust::device_vector<T> Vr(V.size());
  // get real part from device and copy to host
  thrust::transform(V.begin(),V.end(),Vr.begin(),real_functor<T>());
  thrust::copy(Vr.begin(),Vr.end(),pr);
  // get imag part from device and copy to host
  thrust::transform(V.begin(),V.end(),Vr.begin(),imag_functor<T>());
  thrust::copy(Vr.begin(),Vr.end(),pi);  
}

template <typename T, typename Vector> 
void complexCopyHost(mxArray *plhs[],const mxArray *prhs[], mxClassID classid){
  Vector &V=get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  mwSize dims0[1];
  dims0[0]=V.size();
  plhs[0] = mxCreateNumericArray(1,dims0,classid,mxCOMPLEX);
  T *pr= (T *)mxGetPr(plhs[0]);
  T *pi= (T *)mxGetPi(plhs[0]);
  
  // get real part from device and copy to host
  thrust::transform(V.begin(),V.end(),pr,real_functor<T>());
  //thrust::copy(Vr.begin(),Vr.end(),pr);
  // get imag part from device and copy to host
  thrust::transform(V.begin(),V.end(),pi,imag_functor<T>());
  //thrust::copy(Vr.begin(),Vr.end(),pi);  
}

template <typename T, typename Vector> 
void CopyReal(mxArray *plhs[],const mxArray *prhs[], mxClassID classid){
  Vector &V=get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  mwSize dims0[1];
  dims0[0]=V.size();
  plhs[0] = mxCreateNumericArray(1,dims0,classid,mxREAL);
  T *pr= (T *)mxGetPr(plhs[0]);
  
  thrust::copy(V.begin(),V.end(),pr);
}


template <typename T>
void mx_mThrust_gather_type(mxArray *plhs[], const mxArray *prhs[], mxClassID classid){
  char *container=mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  logical *iscomp=(logical *)mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  
  if (strcmp(container, "host")==0){
    typedef thrust::host_vector<thrust::complex<T> > VecCplx;
    typedef thrust::host_vector<T>                   VecReal;
    if (*iscomp)
      complexCopyHost<T,VecCplx>(plhs,prhs,classid);
    else
      CopyReal<T,VecReal>(plhs,prhs,classid);
  }else if (strcmp(container, "device")==0){
    typedef thrust::device_vector<thrust::complex<T> > VecCplx;
    typedef thrust::device_vector<T>                   VecReal;
    if (*iscomp)
      complexCopyDevice<T,VecCplx>(plhs,prhs,classid);
    else
      CopyReal<T,VecReal>(plhs,prhs,classid);
  }else{
    mxFree(container);
    mexErrMsgIdAndTxt("mThrustToolbox:gather","Unknow container."); 
  }
  mxFree(container);
}

// prhs[0] -> Input mThrust object
// plhs[0] -> Output Matlab vector
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=1 || nrhs != 1) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust"));
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  
  if (strcmp(type, "double")==0)
    mx_mThrust_gather_type<double>(plhs,prhs,mxDOUBLE_CLASS);
  else if (strcmp(type, "float")==0)
    mx_mThrust_gather_type<float>(plhs,prhs,mxSINGLE_CLASS);
  else if (strcmp(type, "int")==0)
    mx_mThrust_gather_type<int>(plhs,prhs,mxINT32_CLASS);
  else if (strcmp(type, "unsigned")==0)
    mx_mThrust_gather_type<unsigned>(plhs,prhs,mxUINT32_CLASS);
  else if (strcmp(type, "logical")==0)
    mx_mThrust_gather_type<logical>(plhs,prhs,mxLOGICAL_CLASS);
  else{
    mxFree(type);
    mexErrMsgIdAndTxt("mThrustToolbox:gather","type not implemented.");
  }
  mxFree(type);
  MEXPRINTF("ending");  
}