/*! \file mex_mThrust_plus.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"

#define MESS "mThrustToolbox:jn"


struct op_functor_float : public thrust::unary_function<float,float>
{
  const int n; 
  op_functor_float(int _n) : n(_n) {};
  
  __host__ __device__ 
  float operator()(const float& x) const { 
    return jnf(n,x);
  }
};

struct op_functor_double : public thrust::unary_function<double,double>
{
  const int n; 
  op_functor_double(int _n) : n(_n) {};
  
  __host__ __device__ 
  double operator()(const double& x) const { 
    return jn(n,x);
  }
};


template <typename Vec> 
void mx_mThrust_jn_float_container(const mxArray *prhs[]){
  Vec &gx = get_object< Vec >(mxGetProperty(prhs[1],0,"handle"));
  Vec &gy = get_object< Vec >(mxGetProperty(prhs[2],0,"handle"));
  int *pn = (int *)mxGetPr(prhs[0]);
  
  assert((gx.size()==gy.size()));
  thrust::transform(gx.begin(),gx.end(), gy.begin(), op_functor_float(*pn));
}

template <typename Vec> 
void mx_mThrust_jn_double_container(const mxArray *prhs[]){
  Vec  &gx = get_object< Vec >(mxGetProperty(prhs[1],0,"handle"));
  Vec &gy = get_object< Vec >(mxGetProperty(prhs[2],0,"handle"));
  int *pn = (int *)mxGetPr(prhs[0]);
  
  assert((gx.size()==gy.size()));
  thrust::transform(gx.begin(),gx.end(), gy.begin(), op_functor_double(*pn));
}

void mx_mThrust_jn_float(const mxArray *prhs[]){
  char *container[] = {
    mxArrayToString(mxGetProperty(prhs[1],0,"container")), 
    mxArrayToString(mxGetProperty(prhs[2],0,"container"))
  };
  assert( (strcmp(container[0],container[1])==0) );
  int *pn = (int *)mxGetPr(prhs[0]);
  if (strcmp(container[0], "host")==0){
    typedef thrust::host_vector<float> Vec;
    mx_mThrust_jn_float_container<Vec>(prhs);
  }else if (strcmp(container[0], "device")==0){
    typedef thrust::device_vector<float> Vec;
    mx_mThrust_jn_float_container<Vec>(prhs);
  }else{
    mxFree(container[0]);mxFree(container[1]);
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
  }
  mxFree(container[0]);mxFree(container[1]);
}

void mx_mThrust_jn_double(const mxArray *prhs[]){
  char *container[] = {
    mxArrayToString(mxGetProperty(prhs[1],0,"container")), 
    mxArrayToString(mxGetProperty(prhs[2],0,"container"))
  };
  assert( (strcmp(container[0],container[1])==0) );
  int *pn = (int *)mxGetPr(prhs[0]);
  if (strcmp(container[0], "host")==0){
    typedef thrust::host_vector<double> Vec;
    mx_mThrust_jn_double_container<Vec>(prhs);
  }else if (strcmp(container[0], "device")==0){
    typedef thrust::device_vector<double> Vec;
    mx_mThrust_jn_double_container<Vec>(prhs);
  }else{
    mxFree(container[0]);mxFree(container[1]);
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
  }
  mxFree(container[0]);mxFree(container[1]);
}



// prhs[0] -> order (int)
// prhs[1] -> input  mThrust real (float or double)
// prhs[2] -> output mThrust real (float or double) 
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 3) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"int32"));
  assert(mxIsClass(prhs[1],"mThrust"));
  assert(mxIsClass(prhs[2],"mThrust"));
  
  char *type[] = {
    mxArrayToString(mxGetProperty(prhs[1],0,"type")), 
    mxArrayToString(mxGetProperty(prhs[2],0,"type"))
  };
  assert( (strcmp(type[0],type[1])==0) );
  logical *pc1=(logical *)mxGetPr(mxGetProperty(prhs[1],0,"complex"));
  logical *pc2=(logical *)mxGetPr(mxGetProperty(prhs[2],0,"complex"));
  assert( !(*pc1) && !(*pc2) );
  if (strcmp(type[0], "float")==0){
    mx_mThrust_jn_float(prhs);
  }else if (strcmp(type[0], "double")==0){
    mx_mThrust_jn_double(prhs);
  }else{
    mxFree(type[0]);mxFree(type[1]);
    mexErrMsgIdAndTxt(MESS,"type not implemented.");
  }
  mxFree(type[0]);mxFree(type[1]);
  MEXPRINTF("ending");
}