/*! \file mex_mThrust_plus.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"

#define MESS "mThrustToolbox:cstpower"

// cst_cmplex .^ VectorComplex with T==thrust::complex<...>
// cst_real .^ VectorReal with T==int, double, ... 
template <typename T1,typename T2,typename Tout> struct vecOPcst_functor : public thrust::unary_function<T2,Tout>
{
  const T1 a;
  
  vecOPcst_functor(T1 _a) : a(_a) {}
  
  __host__ __device__
  Tout operator()(const T2& x) const { 
    return pow(x,a);
  }
};

// Vector .^ Cst(real) (ie z=x.^a )
template <typename T, typename VectorIn, typename VectorOut> 
void mx_mThrust_vecOPcstreal(const mxArray *prhs[]){
  typedef typename VectorIn::value_type Tin;
  typedef typename VectorOut::value_type Tout;
  T *pa = (T *)mxGetPr(prhs[1]);
  VectorIn &gx = get_object< VectorIn >(mxGetProperty(prhs[0],0,"handle"));
  VectorOut &gz = get_object< VectorOut >(mxGetProperty(prhs[2],0,"handle"));
  assert((gx.size()==gz.size()));
  thrust::transform(gx.begin(),gx.end(),gz.begin(), vecOPcst_functor<T,Tin,Tout>(*pa));
}
// Vector .^ Cst(complex) (ie z=x.^a )
template <typename T, typename VectorIn, typename VectorOut> 
void mx_mThrust_vecOPcstcplx(const mxArray *prhs[]){
  typedef typename VectorIn::value_type Tin;
  typedef typename VectorOut::value_type Tout;
  T *pa_r = (T *)mxGetPr(prhs[1]);
  T *pa_i = (T *)mxGetPi(prhs[1]);
  thrust::complex<T> a=thrust::complex<T>(*pa_r,*pa_i);
  VectorIn &gx = get_object< VectorIn >(mxGetProperty(prhs[0],0,"handle"));
  VectorOut &gz = get_object< VectorOut >(mxGetProperty(prhs[2],0,"handle"));
  assert((gx.size()==gz.size()));
  thrust::transform(gx.begin(),gx.end(),gz.begin(), vecOPcst_functor<thrust::complex<T>,Tin,Tout>(a));
}

template  <typename T>
void mx_mThrust_vecOPcst_type(const mxArray *prhs[]){
  char *container[] = {
    mxArrayToString(mxGetProperty(prhs[0],0,"container")),
    mxArrayToString(mxGetProperty(prhs[2],0,"container"))
  };
  assert( (strcmp(container[0],container[1])==0) );
  bool iscomplex[3]; 
  iscomplex[1]=mxIsComplex(prhs[1]);
  logical *iscomp=(logical *)mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  iscomplex[0]=*iscomp;
  iscomp=(logical *)mxGetPr(mxGetProperty(prhs[2],0,"complex"));
  iscomplex[2]=*iscomp;
  if (strcmp(container[0], "host")==0){
    typedef typename thrust::host_vector<thrust::complex<T> > VecCplx;
    typedef typename thrust::host_vector<T> VecReal;
    if (iscomplex[0] && iscomplex[1])
      mx_mThrust_vecOPcstcplx< T, VecCplx, VecCplx >(prhs);
    else if (iscomplex[1]) 
      mx_mThrust_vecOPcstcplx< T, VecReal, VecCplx >(prhs);
    else if (iscomplex[0])
      mx_mThrust_vecOPcstreal< T, VecCplx, VecCplx >(prhs);
    else
      mx_mThrust_vecOPcstreal< T, VecReal, VecReal >(prhs);
  }else if (strcmp(container[0], "device")==0){
    typedef typename thrust::device_vector<thrust::complex<T> > VecCplx;
    typedef typename thrust::device_vector<T> VecReal;
    if (iscomplex[0] && iscomplex[1])
      mx_mThrust_vecOPcstcplx< T, VecCplx, VecCplx >(prhs);
    else if (iscomplex[1]) 
      mx_mThrust_vecOPcstcplx< T, VecReal, VecCplx >(prhs);
    else if (iscomplex[0])
      mx_mThrust_vecOPcstreal< T, VecCplx, VecCplx >(prhs);
    else
      mx_mThrust_vecOPcstreal< T, VecReal, VecReal >(prhs);
  }else
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
}

void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 3) 
    mexErrMsgTxt("Bad input.");
  if ( !mxIsScalar(prhs[1]) || !mxIsClass(prhs[0],"mThrust") || !mxIsClass(prhs[2],"mThrust") )
    mexErrMsgIdAndTxt(MESS,"Bad input.");
  
  char *type[] = {
    mxArrayToString(mxGetProperty(prhs[0],0,"type")),
    mxArrayToString(mxGetProperty(prhs[2],0,"type"))
  };
  assert( strcmp(type[0],type[1])==0 );
  if (strcmp(type[0], "float")==0){
    assert( (mxGetClassID(prhs[1])==mxSINGLE_CLASS) );
    mx_mThrust_vecOPcst_type<float>(prhs);
  }else if (strcmp(type[0], "double")==0){
    assert( (mxGetClassID(prhs[1])==mxDOUBLE_CLASS) );
    mx_mThrust_vecOPcst_type<double>(prhs);
  }else{
    mxFree(type[0]);mxFree(type[1]);
    mexErrMsgIdAndTxt(MESS,"type not implemented.");
  }
  MEXPRINTF("ending");
}