/*! \file mex_mThrust_set.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <time.h>


// #define PRINT_DEBUG(C,T,N,A) printf("%sdebug : %s_vector<%s>[%d] (memory adress: %#lX)%s\n",MEXCOLOR,C,T,N,(MEMORY_TYPE)A,TEXTCOLOR_END);

#include <thrust/random.h>
#include <thrust/random/linear_congruential_engine.h>
#include <thrust/random/normal_distribution.h>
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <thrust/transform.h>
#include <thrust/iterator/counting_iterator.h>

__host__ __device__
unsigned int hash(unsigned int a)
{
    a = (a+0x7ed55d16) + (a<<12);
    a = (a^0xc761c23c) ^ (a>>19);
    a = (a+0x165667b1) + (a<<5);
    a = (a+0xd3a2646c) ^ (a<<9);
    a = (a+0xfd7046c5) + (a<<3);
    a = (a^0xb55a4f09) ^ (a>>16);
    return a;
}

template <typename T>
struct prgreal
{
    T a, b;
    unsigned int t;

    __host__ __device__
    prgreal(T _a=0.f, T _b=1.f, unsigned int _t=0 ) : a(_a), b(_b), t(_t) {};

    __host__ __device__
        T operator()(unsigned int thread_id) const
        {
	  unsigned int seed = hash(thread_id+t);
          //thrust::default_random_engine rng;
	  thrust::minstd_rand rng(seed);
          //thrust::random::experimental::normal_distribution<T> dist(a, b);
          // With CUDA 6.0 :
          thrust::random::normal_distribution<T> dist(a, b);
          //  rng.discard(n);

          return dist(rng);
        }
};

// prhs[0] : input/output mThrust object
// prhs[1] : mean
// prhs[2] : standard deviation
// prhs[3] : number of generated values
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 4) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust"));
  assert(mxIsClass(prhs[3],"uint32"));
  
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  char *container =mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  mxArray *mxhandle=mxGetProperty(prhs[0],0,"handle");
  unsigned *seed = (unsigned *) mxGetPr(prhs[3]);
  if( mxGetNumberOfElements(prhs[1])!=1 )
    mexErrMsgIdAndTxt("mThrust::randn","2nd input value must be a scalar.");
  if( mxGetNumberOfElements(prhs[2])!=1 )
    mexErrMsgIdAndTxt("mThrust::randn","3rd input value must be a scalar.");
  
  if (strcmp(type, "int")==0){
    mexErrMsgIdAndTxt("mThrustToolbox:randn","type <int> not allowed.");
  }else if (strcmp(type, "unsigned")==0){
    mexErrMsgIdAndTxt("mThrustToolbox:randn","type <unsigned> not allowed.");
  }else if (strcmp(type, "float")==0){
    float *pM = (float *) mxGetPr(prhs[1]);
    float *pS = (float *) mxGetPr(prhs[2]);
    if (strcmp(container, "host")==0){
      thrust::host_vector<float> &pV=get_object< thrust::host_vector<float> >(mxhandle);
      thrust::transform(thrust::counting_iterator<unsigned int>(0),
                        thrust::counting_iterator<unsigned int>(pV.size()),
                        pV.begin(),
                        prgreal<float>(*pM,*pS,*seed));
    }else if (strcmp(container, "device")==0){
      thrust::device_vector<float> &pV=get_object< thrust::device_vector<float> >(mxhandle);
      thrust::transform(thrust::counting_iterator<unsigned int>(0),
                        thrust::counting_iterator<unsigned int>(pV.size()),
                        pV.begin(),
                        prgreal<float>(*pM,*pS,*seed));
    }else
      mexErrMsgIdAndTxt("mThrustToolbox:randn","container not implemented.");
  }else if (strcmp(type, "double")==0){
    double *pM = (double *) mxGetPr(prhs[1]);
    double *pS = (double *) mxGetPr(prhs[2]);
    if (strcmp(container, "host")==0){
      thrust::host_vector<double> &pV=get_object< thrust::host_vector<double> >(mxhandle);
      thrust::transform(thrust::counting_iterator<unsigned int>(0),
                        thrust::counting_iterator<unsigned int>(pV.size()),
                        pV.begin(),
                        prgreal<double>(*pM,*pS,*seed));
    }else if (strcmp(container, "device")==0){
      thrust::device_vector<double> &pV=get_object< thrust::device_vector<double> >(mxhandle);
      thrust::transform(thrust::counting_iterator<unsigned int>(0),
                        thrust::counting_iterator<unsigned int>(pV.size()),
                        pV.begin(),
                        prgreal<double>(*pM,*pS,*seed));
    }else
      mexErrMsgIdAndTxt("mThrustToolbox:randn","container not implemented.");
  }else  if (strcmp(type, "logical")==0){
    mexErrMsgIdAndTxt("mThrustToolbox:randn","type <logical> not allowed.");
  }else
    mexErrMsgIdAndTxt("mThrustToolbox:randn","type not implemented.");
}