/*! \file mex_mThrust_set.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <thrust/sequence.h>

#define MESS "mThrustToolbox:sequence"

template <typename T,typename Vector> 
void mx_mThrust_sequence(const mxArray *prhs[]){
  Vector &V = get_object< Vector >(mxGetProperty(prhs[0],0,"handle"));
  logical *iscomp=(logical *)mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  if( mxGetNumberOfElements(prhs[1])!=1 )
    mexErrMsgIdAndTxt("mThrust:sequence","2nd input value must be a scalar.");
  if( mxGetNumberOfElements(prhs[2])!=1 )
    mexErrMsgIdAndTxt("mThrust:sequence","3rd input value must be a scalar.");
  if( mxGetNumberOfElements(prhs[3])!=1 )
    mexErrMsgIdAndTxt("mThrust:sequence","4th input value must be a scalar.");
  if( mxGetNumberOfElements(prhs[4])!=1 )
    mexErrMsgIdAndTxt("mThrust:sequence","5th input value must be a scalar.");
  
  int *pstart = (int *) mxGetPr(prhs[1]);
  int *pend = (int *) mxGetPr(prhs[2]);
  if (*iscomp){
    T *pinit_r = (T *) mxGetPr(prhs[3]);
    T *pinit_i = (T *) mxGetPi(prhs[3]);
    T *pstep_r = (T *) mxGetPr(prhs[4]);
    T *pstep_i = (T *) mxGetPi(prhs[4]);
    thrust::complex<T> init(*pinit_r,*pinit_i);
    thrust::complex<T> step(*pstep_r,*pstep_i);
    // thrust::sequence(V.begin()+*pstart,V.begin()+*pend,init,step); // COMPILE FAILED
  }else{
    T *pinit = (T *) mxGetPr(prhs[3]);
    T *pstep = (T *) mxGetPr(prhs[4]);
    thrust::sequence(V.begin()+*pstart,V.begin()+*pend,*pinit,*pstep);
  }
}

template  <typename T>
void mx_mThrust_sequence_type(const mxArray *prhs[]){
  char *container=mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  logical *iscomp=(logical *)mxGetPr(mxGetProperty(prhs[0],0,"complex"));
 
  if (strcmp(container, "device")==0){
    if (*iscomp)
      mx_mThrust_sequence<T,thrust::device_vector<thrust::complex<T> > >(prhs); 
    else
      mx_mThrust_sequence<T,thrust::device_vector<T> >(prhs); 
  }else if (strcmp(container, "host")==0){
    if (*iscomp)
      mx_mThrust_sequence<T,thrust::host_vector<thrust::complex<T> > >(prhs); 
    else
      mx_mThrust_sequence<T,thrust::host_vector<T> >(prhs);  
  }else{
    mxFree(container);
    mexErrMsgIdAndTxt(MESS,"Unknow container."); 
  }
  mxFree(container);
}


// prhs[0] : mThrust object (size dim)
// prhs[1] : int scalar start index (>=0) and (<dim)
// prhs[2] : int scalar end index (>=0) and (<dim)
// prhs[3] : scalar init value
// prhs[4] : scalar step value
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 5) 
    mexErrMsgTxt("Bad input.");
  mxAssert(mxIsClass(prhs[0],"mThrust"),"1st parameter must be <mThrust> object");
  logical *iscomp=(logical *)mxGetPr(mxGetProperty(prhs[0],0,"complex"));
  mxAssert( !*iscomp , "Complex not implemented"); // Owing to failed compilation 
  
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  if (strcmp(type, "double")==0)
    mx_mThrust_sequence_type<double>(prhs);
  else if (strcmp(type, "float")==0)
    mx_mThrust_sequence_type<float>(prhs);
  else if (strcmp(type, "int")==0)
    mx_mThrust_sequence_type<int>(prhs);
  else if (strcmp(type, "unsigned")==0)
    mx_mThrust_sequence_type<unsigned>(prhs);
  else{
    mxFree(type);
    mexErrMsgIdAndTxt("mThrustToolbox:sequence","type not implemented.");
  }
  mxFree(type);
  MEXPRINTF("ending");
}
