/*! \file mex_mThrust_sortIn.cu
 *  \brief MEX-file : Sort in place
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <time.h>

// #define PRINT_DEBUG(C,T,N,A) printf("%sdebug : %s_vector<%s>[%d] (memory adress: %#lX)%s\n",MEXCOLOR,C,T,N,(MEMORY_TYPE)A,TEXTCOLOR_END);

#include <thrust/sort.h>
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>
#include <thrust/functional.h>
#include <thrust/execution_policy.h>

#define MESS "mThrustToolbox:mThrust_sort"

template <typename T,typename Vector> void  mThrust_sort(
  const mxArray *mx_mThrust,unsigned Flag){
  
  Vector &V = get_object< Vector >(mx_mThrust);
  if (Flag==0)
    thrust::sort(V.begin(),V.end(),thrust::greater<T>());
  else
    //thrust::sort(V.begin(),V.end(),thrust::greater<T>());
    thrust::sort(V.begin(),V.end(),thrust::less<T>());
}

// prhs[0] -> mThrust object to sort
// prhs[1] -> double, 0 for ascending order otherwise descending order
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 2) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust") && (mxGetClassID(prhs[1])==mxUINT32_CLASS));
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  char *container =mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  mxArray *mxhandle=mxGetProperty(prhs[0],0,"handle");
  unsigned *pFlag = (unsigned *)mxGetData(prhs[1]);

  if (strcmp(container, "host")==0){
    if (strcmp(type, "int")==0){
      mThrust_sort<int,thrust::host_vector<int> >(mxhandle,*pFlag);
    }else if (strcmp(type, "double")==0){
      mThrust_sort<double,thrust::host_vector<double> >(mxhandle,*pFlag);
    }else if (strcmp(type, "float")==0){
      mThrust_sort<float,thrust::host_vector<float> >(mxhandle,*pFlag);
    }else if (strcmp(type, "unsigned")==0){
      mThrust_sort<unsigned,thrust::host_vector<unsigned> >(mxhandle,*pFlag);
    }else if (strcmp(type, "logical")==0){
      mThrust_sort<logical,thrust::host_vector<logical> >(mxhandle,*pFlag);
    }else
      mexWarnMsgIdAndTxt(MESS,"Type %s not allowed.",type);
  }else if (strcmp(container, "device")==0){
    if (strcmp(type, "int")==0){
      mThrust_sort<int,thrust::device_vector<int> >(mxhandle,*pFlag);
    }else if (strcmp(type, "double")==0){
      mThrust_sort<double,thrust::device_vector<double> >(mxhandle,*pFlag);
    }else if (strcmp(type, "float")==0){
      mThrust_sort<float,thrust::device_vector<float> >(mxhandle,*pFlag);
    }else if (strcmp(type, "unsigned")==0){
      mThrust_sort<unsigned,thrust::device_vector<unsigned> >(mxhandle,*pFlag);
    }else if (strcmp(type, "logical")==0){
      mThrust_sort<logical,thrust::device_vector<logical> >(mxhandle,*pFlag);
    }else
      mexWarnMsgIdAndTxt(MESS,"Type %s not allowed.",type);
  }else
    mexWarnMsgIdAndTxt(MESS,"Container %s not allowed.",container);
  mxFree(type);mxFree(container);
  MEXPRINTF("ending");
}