/*! \file mex_mThrust_sortIn.cu
 *  \brief MEX-file : Sort in place
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"
#include <time.h>

// #define PRINT_DEBUG(C,T,N,A) printf("%sdebug : %s_vector<%s>[%d] (memory adress: %#lX)%s\n",MEXCOLOR,C,T,N,(MEMORY_TYPE)A,TEXTCOLOR_END);

#include <thrust/sort.h>
#include <thrust/host_vector.h>
#include <thrust/device_vector.h>

#define MESS "mThrustToolbox:sort_with_index"

template <typename T,typename Vector,typename VectorInd> void  mThrust_sort_by_key(
  const mxArray *mx_keys,const mxArray *mx_index, unsigned Flag){
  
  Vector &keys = get_object< Vector >(mx_keys);
  VectorInd &Index = get_object< VectorInd >(mx_index);
  if (Flag==0)
    thrust::sort_by_key(keys.begin(),keys.end(),Index.begin(),thrust::greater<T>());
  else
    thrust::sort_by_key(keys.begin(),keys.end(),Index.begin(),thrust::less<T>());
  }

// prhs[0] -> keys, mThrust vector
// prhs[1] -> index, mThrust vector same container as keys and type==int)
// prhs[2] -> double, 0 for ascending order otherwise descending order
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 3) 
    mexErrMsgTxt("Bad input.");
  assert(mxIsClass(prhs[0],"mThrust") && mxIsClass(prhs[1],"mThrust") && (mxGetClassID(prhs[2])==mxUINT32_CLASS));
  char *type =mxArrayToString(mxGetProperty(prhs[0],0,"type"));
  char *container =mxArrayToString(mxGetProperty(prhs[0],0,"container"));
  mxArray *mxhandle_keys=mxGetProperty(prhs[0],0,"handle");
  mxArray *mxhandle_values=mxGetProperty(prhs[1],0,"handle");
  unsigned *pFlag = (unsigned *)mxGetData(prhs[2]);
  
  if (strcmp(container, "host")==0){
    //thrust::host_vector<int> &index=get_object< thrust::host_vector<int> >(mxhandle_values);
    if (strcmp(type, "float")==0){
      mThrust_sort_by_key<float,thrust::host_vector<float>,thrust::host_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "double")==0){
      mThrust_sort_by_key<double,thrust::host_vector<double>,thrust::host_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "int")==0){
      mThrust_sort_by_key<int,thrust::host_vector<int>,thrust::host_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "unsigned")==0){
      mThrust_sort_by_key<unsigned,thrust::host_vector<unsigned>,thrust::host_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "logical")==0){
      mThrust_sort_by_key<logical,thrust::host_vector<logical>,thrust::host_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else
      mexWarnMsgIdAndTxt(MESS,"SortWithIndex of mThrust<%s> not allowed.",type);
  }else if (strcmp(container, "device")==0){  
    //thrust::device_vector<int> &index=get_object< thrust::device_vector<int> >(mxhandle_values);
    if (strcmp(type, "float")==0){
      mThrust_sort_by_key<float,thrust::device_vector<float>,thrust::device_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "double")==0){
      mThrust_sort_by_key<double,thrust::device_vector<double>,thrust::device_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "int")==0){
      mThrust_sort_by_key<int,thrust::device_vector<int>,thrust::device_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "unsigned")==0){
      mThrust_sort_by_key<unsigned,thrust::device_vector<unsigned>,thrust::device_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else if (strcmp(type, "logical")==0){
      mThrust_sort_by_key<logical,thrust::device_vector<logical>,thrust::device_vector<int> >(mxhandle_keys,mxhandle_values,*pFlag);
    }else
      mexWarnMsgIdAndTxt(MESS,"SortWithIndex of mThrust<%s> not allowed.",type);
  }else
    mexWarnMsgIdAndTxt(MESS,"container '%s' not allowed.",container);
  mxFree(type);mxFree(container);
  MEXPRINTF("ending");
}