/*! \file mex_mThrust_plus.cu
 *  \brief MEX-file : delete a C++ \p gTArray object (GPU).
 * 
 *  The gateway Cuda/C++ function \p mex_gTArrayDestroy destroy a C++ \p gTArray object 
 *  pointed by a handle.
 * 
 *  &copy;Matlab syntax <tt><B>mex_gTArrayDelete(h,type)</B></tt> where
 *    - \p h  handle representing a C++ \p gTArray object.
 *    - \p type of \p gTArray object given by \p h. Must be 'int', 'unsigned', 'float' or 'double'.
 * 
 *  The following &copy;Matlab code snippet demonstrates how to use this gateway function \p mex_gTArrayDelete 
 *  \code
 *  >> hA=mex_gTArrayCreate(int32(500),'double');
 *  ...
 *  >> mex_gTArrayDelete(hA,'double');
 *  \endcode
 * 
 * \sa  mex_gTArrayCreate
 */
#include <string>
#include "textcolors.h"
#include "ObjectHandle.h"
#include "mThrust.h"
#include "mex.h"

template <typename T> struct functor1 : public thrust::binary_function<T,T,T>
{
  __host__ __device__
  T operator()(const T & z1, const T& z2) const { 
    return z1*z1+z2;
  }
};

template <typename T> struct functor2 : public thrust::binary_function<T,thrust::complex<T>,T>
{
  __host__ __device__
  T operator()(const T & count, const thrust::complex<T>& z) const { 
    return count+(abs(z)<=2?1:0);
  }
};

// prhs[0] : z0 mThrust  complex double
// prhs[1] : count mThrust  double
// prhs[2] : maxIter (int)
void mexFunction( int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[] )     
{
  MEXPRINTF("starting");
  if (nlhs !=0 || nrhs != 3) 
    mexErrMsgTxt("Bad input.");
  typedef thrust::device_vector<thrust::complex<double> > VecCplx;
  typedef thrust::device_vector<double>                   VecReal;
  
  VecCplx &z0    = get_object< VecCplx >(mxGetProperty(prhs[0],0,"handle"));
  VecReal &count = get_object< VecReal >(mxGetProperty(prhs[1],0,"handle"));
  assert( (mxGetClassID(prhs[2])==mxINT32_CLASS) );
  int *pmaxIter = (int *) mxGetPr(prhs[2]);
  VecCplx z=z0;
  
  for (int n=0;n<=*pmaxIter;n++){
    thrust::transform(z.begin(),z.end(), z0.begin(),z.begin(), functor1<thrust::complex<double> >());
    thrust::transform(count.begin(),count.end(), z.begin(),count.begin(), functor2<double>());
  }
  
  MEXPRINTF("ending");
}